"""Class implementation for the timer.
"""

from typing import Any
from typing import Callable
from typing import Optional
from typing import TypeVar
from typing import Union

from apysc._event import timer_event
from apysc._event.custom_event_interface import CustomEventInterface
from apysc._event.handler import HandlerData
from apysc._time.fps import FPS
from apysc._type.boolean import Boolean
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.number_value_interface import NumberValueInterface
from apysc._type.variable_name_interface import VariableNameInterface

_O1 = TypeVar('_O1')
_O2 = TypeVar('_O2')
_Handler = Callable[['timer_event.TimerEvent', _O1], None]


class Timer(VariableNameInterface, CustomEventInterface):
    """
    Timer class to handle function calling at regular intervals.

    References
    ----------
    - Timer document
        - https://simon-ritchie.github.io/apysc/timer.html
    - TimerEvent class document
        - https://simon-ritchie.github.io/apysc/timer_event.html
    - Timer class delay setting document
        - https://simon-ritchie.github.io/apysc/timer_delay.html
    - FPS enum document
        - https://simon-ritchie.github.io/apysc/fps.html
    - Timer class repeat_count setting
        - https://simon-ritchie.github.io/apysc/timer_repeat_count.html
    """

    _delay: Number
    _repeat_count: Int
    _current_count: Int
    _handler_data: HandlerData
    _handler_name: str
    _running: Boolean

    def __init__(
            self,
            handler: _Handler[_O1],
            delay: Union[int, float, NumberValueInterface, FPS],
            *,
            repeat_count: Union[int, Int] = 0,
            options: Optional[_O1] = None) -> None:
        """
        Timer class to handle function calling at regular intervals.

        Parameters
        ----------
        handler : _Handler
            A handler would be called at regular intervals.
        delay : int or float or Int or Number or FPS
            A delay between each handler's calling in milisecond
            or FPS value.
            If `FPS` value is specified, then this value will be
            a milisecond calculated with that FPS value (e.g., if
            the `FPS_60` value is specified, then `delay` will be
            16.6666667).
        repeat_count : int or Int
            Max count of a handler's calling. If the handler's calling
            count has reached this value, then a timer will stop.
            If 0 is specified, then a timer will loop forever.
        options : dict or None, default None
            Optional arguments dictionary to pass the the handler.

        References
        ----------
        - Timer document
            - https://simon-ritchie.github.io/apysc/timer.html
        - TimerEvent class document
            - https://simon-ritchie.github.io/apysc/timer_event.html
        - Timer class delay setting document
            - https://simon-ritchie.github.io/apysc/timer_delay.html
        - FPS enum document
            - https://simon-ritchie.github.io/apysc/fps.html
        - Timer class repeat_count setting
            - https://simon-ritchie.github.io/apysc/timer_repeat_count.html
        - About the handler options’ type document
            - https://bit.ly/39tnYxC
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._event.handler import get_handler_name
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._expression.event_handler_scope import \
                TemporaryNotHandlerScope
            from apysc._validation import number_validation
            from apysc._validation.handler_options_validation import \
                validate_options_type
            with TemporaryNotHandlerScope():
                validate_options_type(options=options)
                self.variable_name = \
                    expression_variables_util.get_next_variable_name(
                        type_name=var_names.TIMER)
                self._handler_name = get_handler_name(
                    handler=handler, instance=self)
                handler = self._wrap_handler(handler=handler)
                self._handler: _Handler[_O1] = handler
                delay = self._convert_delay_to_number(delay=delay)
                number_validation.validate_num_is_gte_zero(num=delay)
                self._delay = delay
                if not isinstance(repeat_count, ap.Int):
                    repeat_count = ap.Int(repeat_count)
                number_validation.validate_num_is_gte_zero(num=repeat_count)
                self._repeat_count = repeat_count
                self._running = ap.Boolean(False)
                self._current_count = ap.Int(0)
                if options is None:
                    options = {}  # type: ignore
                self._handler_data = {  # type: ignore
                    'handler': self._handler,  # type: ignore
                    'options': options,
                }
                ap.append_js_expression(
                    expression=f'var {self.variable_name};')

    def _convert_delay_to_number(
            self, *,
            delay: Union[int, float, NumberValueInterface, FPS]) -> Number:
        """
        Convert each type of delay value to a `Number` value.

        Parameters
        ----------
        delay : int or float or Int or Number or FPS
            A delay between each handler's calling in milisecond
            or FPS value.

        Returns
        -------
        delay : Number
            Converted delay value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._convert_delay_to_number, locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._time.fps import FPSDefinition
            if isinstance(delay, FPS):
                fps_definition: FPSDefinition = delay.value
                delay = fps_definition.milisecond_intervals
            if not isinstance(delay, ap.Number):
                delay = ap.Number(delay)
            return delay

    @property
    def delay(self) -> Number:
        """
        Get a delay value.

        Returns
        -------
        delay : Number
            A delay value of each handler's calling in milisecond.

        References
        ----------
        - Timer class delay setting document
            - https://simon-ritchie.github.io/apysc/timer_delay.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='delay', locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._type import value_util
            return value_util.get_copy(value=self._delay)

    @property
    def repeat_count(self) -> Int:
        """
        Get a max count value of a handler's calling.

        Returns
        -------
        repeat_count : Int
            Max count of a handler's calling. If 0 is set, then a
            timer will loop forever.

        References
        ----------
        - Timer class repeat_count setting
            - https://simon-ritchie.github.io/apysc/timer_repeat_count.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='repeat_count', locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._type import value_util
            return value_util.get_copy(value=self._repeat_count)

    @property
    def running(self) -> Boolean:
        """
        Get a boolean value whether this timer is running or not.

        Returns
        -------
        running : Boolean.
            A boolean value whether this timer is running or not.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='running', locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._type import value_util
            return value_util.get_copy(value=self._running)

    @property
    def current_count(self) -> Int:
        """
        Get a current timer count.

        Returns
        -------
        current_count : Int
            A current timer count.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='current_count', locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._type import value_util
            return value_util.get_copy(value=self._current_count)

    def start(self) -> None:
        """
        Start this timer.

        References
        ----------
        - Timer class start and stop interfaces
            - https://simon-ritchie.github.io/apysc/timer_start_and_stop.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.start, locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._event import handler_circular_calling_util
            from apysc._event.handler import append_handler_expression
            from apysc._expression import expression_data_util
            from apysc._type import value_util
            delay_val_str: str = value_util.get_value_str_for_expression(
                value=self._delay)
            is_handler_circular_calling: bool = handler_circular_calling_util.\
                is_handler_circular_calling(handler_name=self._handler_name)
            if is_handler_circular_calling:
                handler_name: str = handler_circular_calling_util.\
                    get_prev_handler_name(handler_name=self._handler_name)
                variable_name: str = handler_circular_calling_util.\
                    get_prev_variable_name(handler_name=self._handler_name)
            else:
                handler_name = self._handler_name
                variable_name = self.variable_name
            expression: str = (
                f'\nif (_.isUndefined({variable_name})) {{'
                f'\n  {variable_name} = setInterval('
                f'{handler_name}, {delay_val_str});'
                '\n}'
            )
            expression_data_util.append_js_expression(expression=expression)

            e: ap.TimerEvent = ap.TimerEvent(this=self)
            append_handler_expression(
                handler_data=self._handler_data,
                handler_name=handler_name,
                e=e)
            self._running.value = True

    def _wrap_handler(self, *, handler: _Handler[_O1]) -> _Handler[_O1]:
        """
        Wrap a handler to update a current count value when
        it is called.

        Parameters
        ----------
        handler : _Handler
            Target handler.

        Returns
        -------
        wrapped : _Handler
            Wrapped handler.
        """
        from apysc._event.timer_event import TimerEvent

        def wrapped(e: TimerEvent, options: Any) -> None:
            """
            Wrapped handler.

            Parameters
            ----------
            e : TimerEvent
                Event instance.
            options : dict
                Optional arguments dictionary.
            """
            e.this._current_count += 1
            e.this._current_count._value = 0
            handler(e, options)
            self._append_count_branch_expression()

        return wrapped

    def _append_count_branch_expression(self) -> None:
        """
        Append the timer stopping expression by the counting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_count_branch_expression,
                locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._event.custom_event_type import CustomEventType
            from apysc._expression import expression_data_util
            from apysc._expression.indent_num import Indent
            from apysc._type import value_util
            current_count_val_str: str = value_util.\
                get_value_str_for_expression(value=self._current_count)
            repeat_count_val_str: str = value_util.\
                get_value_str_for_expression(value=self._repeat_count)
            expression: str = (
                f'if ({repeat_count_val_str} !== 0 '
                f'&& {current_count_val_str} === {repeat_count_val_str}) {{'
                f'\n{self._get_stop_expression(indent_num=1)}'
            )
            expression_data_util.append_js_expression(expression=expression)
            with Indent():
                self._running.value = False
                self.trigger_custom_event(
                    custom_event_type=CustomEventType.TIMER_COMPLETE)
                self.unbind_custom_event_all(
                    custom_event_type=CustomEventType.TIMER_COMPLETE)
                self._current_count.value = 0
            expression = '\n}'
            expression_data_util.append_js_expression(expression=expression)

    def stop(self) -> None:
        """
        Stop this timer.

        - Timer class start and stop interfaces
            - https://simon-ritchie.github.io/apysc/timer_start_and_stop.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.stop, locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._expression import expression_data_util
            self._running.value = False
            expression: str = self._get_stop_expression(indent_num=0)
            expression_data_util.append_js_expression(expression=expression)

    def _get_stop_expression(self, *, indent_num: int) -> str:
        """
        Get a stop interface expression string.

        Parameters
        ----------
        indent_num : int
            Indentation number to append.

        Returns
        -------
        expression : str
            Stop interface expression string.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._get_stop_expression, locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._string import indent_util
            expression: str = (
                f'if (!_.isUndefined({self.variable_name})) {{'
                f'\n  clearInterval({self.variable_name});'
                f'\n  {self.variable_name} = undefined;'
                '\n}'
            )
            expression = indent_util.append_spaces_to_expression(
                expression=expression, indent_num=indent_num)
            return expression

    def reset(self) -> None:
        """
        Reset the timer count and stop this timer.

        References
        ----------
        - Timer class reset interface document
            - https://simon-ritchie.github.io/apysc/timer_reset.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.reset, locals_=locals(),
                module_name=__name__, class_=Timer):
            self.stop()
            self._current_count.value = 0

    def timer_complete(
            self, handler: _Handler[_O2], *,
            options: Optional[_O2] = None) -> str:
        """
        Add a timer complete event listener setting.

        Parameters
        ----------
        handler : _Handler
            A callable will be called when a timer is complete.
        options : dict or None, default None
            Optional arguments dictionary to be passed to a handler.

        Returns
        -------
        name : str
            Handler's name.

        References
        ----------
        - Timer class timer_complete interface document
            - https://simon-ritchie.github.io/apysc/timer_complete.html
        - About the handler options’ type document
            - https://bit.ly/39tnYxC
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.timer_complete, locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._event.custom_event_type import CustomEventType
            from apysc._validation.handler_options_validation import \
                validate_options_type
            validate_options_type(options=options)
            e: ap.TimerEvent = ap.TimerEvent(this=self)
            name: str = self.bind_custom_event(
                custom_event_type=CustomEventType.TIMER_COMPLETE,
                handler=handler,
                e=e,
                options=options)
            return name



"""Common callable utility implementations.
"""

import inspect
from inspect import Parameter
from inspect import Signature
from typing import Any
from typing import Callable
from typing import Dict

empty = inspect.Signature.empty


def get_func_default_vals(*, func: Callable) -> Dict[str, Any]:
    """
    Get specified function's arguments default values.

    Parameters
    ----------
    func : Callable
        Target function (or method).

    Returns
    -------
    default_vals : dict
        Dictionary that has a argument name at key and default value
        at value. If argument has no default value, then `empty` object
        will be set.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=get_func_default_vals, locals_=locals(),
            module_name=__name__):
        signature: Signature = inspect.signature(func)
        default_vals: Dict[str, Any] = {}
        parameter: Parameter
        for arg_name, parameter in signature.parameters.items():
            default_vals[arg_name] = parameter.default
        return default_vals


def get_arg_name_at(*, func: Callable, index: int) -> str:
    """
    Get specified index argument name from function.

    Parameters
    ----------
    func : Callable
        Target function (or method).
    index : int
        Argument index (start from 0).

    Returns
    -------
    arg_name : str
        Target index's argument name.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=get_arg_name_at, locals_=locals(),
            module_name=__name__):
        signature: Signature = inspect.signature(func)
        arg_name: str = ''
        for i, target_idx_arg_name in enumerate(signature.parameters.keys()):
            if i != index:
                continue
            arg_name = target_idx_arg_name
            break
        return arg_name


def get_name_and_arg_value_dict_from_args(
        *, func: Callable, args: list, kwargs: dict) -> Dict[str, Any]:
    """
    Get dictionary that has an argument name at key and specified
    argument values at value.

    Parameters
    ----------
    func : Callable
        Target function (or method).
    args : list
        Specified positional arguments.
    kwargs : dict
        Specified keyword arguments.

    Returns
    -------
    name_and_arg_value_dict : dict
        Dictionary that has an argument name at key and specified
        argument values at value.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=get_name_and_arg_value_dict_from_args, locals_=locals(),
            module_name=__name__):
        name_and_arg_value_dict: Dict[str, Any] = {}
        for i, arg_value in enumerate(args):
            arg_name: str = get_arg_name_at(func=func, index=i)
            name_and_arg_value_dict[arg_name] = arg_value
        name_and_arg_value_dict.update(kwargs)
        return name_and_arg_value_dict


def get_method_class_name(*, method: Callable) -> str:
    """
    Get a specified method's class name.

    Parameters
    ----------
    method : Callable
        Target method.

    Returns
    -------
    class_name : str
        Target method's class name. If specified argument is
        not a method (e.g., function), then blank string will be
        returned.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=get_method_class_name, locals_=locals(),
            module_name=__name__):
        if not inspect.ismethod(method):
            return ''
        class_name: str = ''
        for class_ in inspect.getmro(type(method.__self__)):  # type: ignore
            if method.__name__ not in class_.__dict__:
                continue
            class_name = class_.__name__
        return class_name



"""Class implementation for the JavaScript expression string.
"""

from typing import Dict

from apysc._type.revert_interface import RevertInterface


class ExpressionString(RevertInterface):
    """
    The class for the JavaScript expression string.
    """

    _value: str

    def __init__(self, value: str) -> None:
        """
        The class for the JavaScript expression string.

        Parameters
        ----------
        value : str
            String to set.
        """
        self._value = value

    _value_snapshots: Dict[str, str]

    @property
    def value(self) -> str:
        """
        Get a expression string.

        Returns
        -------
        value : str
            A expression string.
        """
        return self._value

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._set_single_snapshot_val_to_dict(
            dict_name='_value_snapshots',
            value=self._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._value = self._value_snapshots[snapshot_name]



"""Class implementation for the parallel animation value.
"""

import re
from typing import Generic
from typing import List
from typing import Match
from typing import Optional
from typing import TypeVar
from typing import Union

from apysc._animation.animation_base import AnimationBase
from apysc._animation.easing import Easing
from apysc._type.int import Int
from apysc._type.variable_name_interface import VariableNameInterface

_T = TypeVar('_T', bound=VariableNameInterface)


class AnimationParallel(AnimationBase[_T], Generic[_T]):
    """
    The parallel animation setting class.
    """

    _animations: List[AnimationBase]

    def __init__(
            self,
            *,
            target: _T,
            animations: List[AnimationBase],
            duration: Union[int, Int] = 3000,
            delay: Union[int, Int] = 0,
            easing: Easing = Easing.LINEAR) -> None:
        """
        The parallel animation setting class.

        Raises
        ------
        ValueError
            - If the animations's target is not unified.
            - If there are changed `duration`, `delay`, or
                `easing` animation settings in the `animations`
                list.

        Parameters
        ----------
        target : VariableNameInterface
            A target instance of the animation target
            (e.g., `DisplayObject` instance).
        animations : list of AnimationBase
            Target animations (e.g., `AnimationX`, `AnimationFillColor`).
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=AnimationParallel):
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            variable_name: str = expression_variables_util.\
                get_next_variable_name(type_name=var_names.ANIMATION_PARALLEL)
            self._animations = animations
            self._set_basic_animation_settings(
                target=target,
                duration=duration,
                delay=delay,
                easing=easing)
            super(
                AnimationParallel,
                self).__init__(
                variable_name=variable_name)
            self._validate_animation_targets_are_unified()
            self._validate_animations_duration_are_default_vals()
            self._validate_animations_delay_are_default_vals()
            self._validate_animations_easing_are_default_vals()

    def _validate_animations_easing_are_default_vals(self) -> None:
        """
        Validate whether the animations easing settings are
        default values (not changed).

        Raises
        ------
        ValueError
            If there is an animation target that is changed
            easing setting.
        """
        for animation in self._animations:
            if animation._easing == Easing.LINEAR:
                continue
            err_msg: str = (
                'There is an animation target that is changed '
                f'easing setting: {animation._easing.name}'
                '\nThe easing setting of animation in the `animations` '
                'argument can not be changed.'
                f'\nTarget animation type: {type(animation)}'
            )
            raise ValueError(err_msg)

    def _validate_animations_delay_are_default_vals(self) -> None:
        """
        Validate whether the animations delay settings are
        default values (not changed).

        Raises
        ------
        ValueError
            If there is an animation target that is changed
            delay setting.
        """
        for animation in self._animations:
            if animation._delay._value == 0:
                continue
            err_msg: str = (
                'There is an animation target that is changed '
                f'delay setting: {animation._delay._value}'
                '\nThe delay setting of animation in the `animations`'
                'argument can not be changed.'
                f'\nTarget animation type: {type(animation)}'
            )
            raise ValueError(err_msg)

    def _validate_animations_duration_are_default_vals(self) -> None:
        """
        Validate whether the animations duration settings are
        default values (not changed).

        Raises
        ------
        ValueError
            If there is an animation target that is changed
            duration setting.
        """
        for animation in self._animations:
            if animation._duration._value == 3000:
                continue
            err_msg: str = (
                'There is an animation target that is changed '
                f'duration setting: {animation._duration._value}'
                '\nThe duration setting of animation in the '
                '`animations` argument can not be changed.'
                f'\nTarget animation type: {type(animation)}'
            )
            raise ValueError(err_msg)

    def _validate_animation_targets_are_unified(self) -> None:
        """
        Validate whether the specified animation's targets are unified.

        Raises
        ------
        ValueError
            If the specified animation targets are not unified.
        """
        for animation in self._animations:
            if animation._target == self._target:
                continue
            err_msg: str = (
                'There is not unified animation target instance: '
                f'{animation._target} (type: {type(animation._target)})'
                f'\nExpected instance: {self._target} (type: '
                f'{type(self._target)})'
                f'\nTarget animation type: {type(animation)}'
                f'\nPlease unify the animation targets of the '
                'animation_parallel interface argument.'
            )
            raise ValueError(err_msg)

    def _get_animation_func_expression(self) -> str:
        """
        Get a animation function expression.

        Returns
        -------
        expression : str
            Animation function expression.
        """
        expression: str = ''
        attr_strs: List[str] = []
        for i, animation in enumerate(self._animations):
            single_expression: str = \
                animation._get_animation_func_expression()
            single_expression = single_expression.replace(';', '')
            if '.attr' in single_expression:
                match: Optional[Match] = re.search(
                    pattern=(
                        r'.*\.attr\({(.*?)}\)'
                    ),
                    string=single_expression,
                    flags=re.MULTILINE | re.DOTALL)
                if match is not None:
                    attr_strs.append(match.group(1))
            else:
                expression += single_expression
            if i == len(self._animations) - 1:
                expression += self._make_animation_attr_exp(
                    attr_strs=attr_strs)
                expression += ';'
        return expression

    def _make_animation_attr_exp(self, *, attr_strs: List[str]) -> str:
        """
        Make an animation attribute expression string.

        Parameters
        ----------
        attr_strs : list of str
            Target attribute strings.

        Returns
        -------
        expression : str
            Created attribute expression string.
        """
        expression: str = '\n  .attr({'
        for i, attr_str in enumerate(attr_strs):
            expression += f'\n    {attr_str}'
            if i != len(attr_strs) - 1:
                expression += ','
        expression += '\n  })'
        return expression

    def _get_complete_event_in_handler_head_expression(self) -> str:
        """
        Get an expression to be inserted into the complete event
        handler's head.

        Returns
        -------
        expression : str
            An expression to be inserted into the complete event
            handler's head.
        """
        expression: str = ''
        for animation in self._animations:
            if expression != '':
                expression += '\n'
            single_expression: str = animation.\
                _get_complete_event_in_handler_head_expression()
            expression += single_expression
        return expression



"""Class implementation for the scale-x from the given point
animation value.
"""

from typing import Generic
from typing import TypeVar
from typing import Union

from apysc._animation.animation_base import AnimationBase
from apysc._animation.easing import Easing
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.variable_name_interface import VariableNameInterface

_T = TypeVar('_T', bound=VariableNameInterface)


class AnimationScaleXFromPoint(AnimationBase[_T], Generic[_T]):
    """
    The animation class for a scale-x from the given point.
    """

    _scale_x_from_point: Number
    _x: Int
    _before_scale_x_from_point: Number
    _scale_x_from_point_diff_ratio: Number

    def __init__(
            self,
            *,
            target: _T,
            scale_x_from_point: Union[float, Number],
            x: Union[int, Int],
            duration: Union[int, Int] = 3000,
            delay: Union[int, Int] = 0,
            easing: Easing = Easing.LINEAR) -> None:
        """
        The animation class for a scale-x from the given point.

        Parameters
        ----------
        target : ScaleXFromPointInterface
            A target instance of the animation target
            (e.g., `Rectangle` instance).
        scale_x_from_point : float or Number
            The final scale-x from the given point of the animation.
        x : int or Int
            X-coordinate.
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Raises
        ------
        TypeError
            If a specified target is not a ScaleXFromPointInterface
            instance.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=AnimationScaleXFromPoint):
            from apysc._converter import to_apysc_val_from_builtin
            from apysc._display.scale_x_from_point_interface import \
                ScaleXFromPointInterface
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            variable_name: str = expression_variables_util.\
                get_next_variable_name(
                    type_name=var_names.ANIMATION_SCALE_X_FROM_POINT)
            self._x = to_apysc_val_from_builtin.\
                get_copied_int_from_builtin_val(integer=x)
            target_: VariableNameInterface = target
            if isinstance(target_, ScaleXFromPointInterface):
                target_._initialize_scale_x_from_point_if_not_initialized()
                self._before_scale_x_from_point = target_.\
                    get_scale_x_from_point(x=self._x)
            else:
                raise TypeError(
                    'Specified `target` argument is not a '
                    f'ScaleXFromPointInterface: {type(target_)}')
            self._scale_x_from_point = to_apysc_val_from_builtin.\
                get_copied_number_from_builtin_val(
                    float_or_num=scale_x_from_point)
            one: ap.Number = ap.Number(1.0)
            self._scale_x_from_point_diff_ratio = (
                one / self._before_scale_x_from_point
                * self._scale_x_from_point)
            self._set_basic_animation_settings(
                target=target,
                duration=duration,
                delay=delay,
                easing=easing)
            super(AnimationScaleXFromPoint, self).__init__(
                variable_name=variable_name)

    def _get_animation_func_expression(self) -> str:
        """
        Get a animation function expression.

        Returns
        -------
        expression : str
            Animation function expression.
        """
        from apysc._type import value_util
        diff_ratio_str: str = value_util.get_value_str_for_expression(
            value=self._scale_x_from_point_diff_ratio)
        x_str: str = value_util.get_value_str_for_expression(value=self._x)
        return f'\n  .scale({diff_ratio_str}, 1, {x_str}, 0);'

    def _get_complete_event_in_handler_head_expression(self) -> str:
        """
        Get an expression to be inserted into the complete event
        handler's head.

        Returns
        -------
        expression : str
            An expression to be inserted into the complete event
            handler's head.
        """
        from apysc._display import scale_interface_helper
        from apysc._display.scale_x_from_point_interface import \
            ScaleXFromPointInterface
        from apysc._type import value_util
        expression: str = ''
        if isinstance(self._target, ScaleXFromPointInterface):
            self._target._initialize_scale_x_from_point_if_not_initialized()
            key_exp_str: str = scale_interface_helper.\
                get_coordinate_key_for_expression(coordinate=self._x).value
            target_scale_x_str: str = value_util.get_value_str_for_expression(
                value=self._target._scale_x_from_point)
            scale_x_str: str = value_util.get_value_str_for_expression(
                value=self._scale_x_from_point)
            expression = (
                f'{target_scale_x_str}[{key_exp_str}] = '
                f'{scale_x_str};'
            )
        return expression



"""Class implementation for the scale_y_from_center interface.
"""

from typing import Dict

from apysc._animation.animation_scale_y_from_center_interface import \
    AnimationScaleYFromCenterInterface
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.number import Number
from apysc._type.revert_interface import RevertInterface


class ScaleYFromCenterInterface(
        AnimationScaleYFromCenterInterface, RevertInterface,
        AttrLinkingInterface):

    _scale_y_from_center: Number

    def _initialize_scale_y_from_center_if_not_initialized(self) -> None:
        """
        Initialize the `_scale_y_from_center` attribute if it hasn't been
        initialized yet.'
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.
                _initialize_scale_y_from_center_if_not_initialized,
                locals_=locals(),
                module_name=__name__, class_=ScaleYFromCenterInterface):
            if hasattr(self, '_scale_y_from_center'):
                return
            self._scale_y_from_center = ap.Number(1.0)

            self._append_scale_y_from_center_attr_linking_setting()

    def _append_scale_y_from_center_attr_linking_setting(self) -> None:
        """
        Append a scale-y attribute linking setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.
                _initialize_scale_y_from_center_if_not_initialized,
                locals_=locals(),
                module_name=__name__, class_=ScaleYFromCenterInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._scale_y_from_center,
                attr_name='scale_y_from_center')
            self._append_attr_to_linking_stack(
                attr=self._scale_y_from_center,
                attr_name='scale_y_from_center')

    @property
    def scale_y_from_center(self) -> Number:
        """
        Get a scale-y value from the center of this instance.

        Returns
        -------
        scale_y_from_center : ap.Number
            Scale-y value from the center of this instance.

        References
        ----------
        - GraphicsBase scale_x_from_center and scale_y_from_center interfaces
            - https://bit.ly/3ityoCX
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='scale_y_from_center', locals_=locals(),
                module_name=__name__, class_=ScaleYFromCenterInterface):
            from apysc._type import value_util
            self._initialize_scale_y_from_center_if_not_initialized()
            return value_util.get_copy(value=self._scale_y_from_center)

    @scale_y_from_center.setter
    def scale_y_from_center(self, value: Number) -> None:
        """
        Update a scale-y value from the center of this instance.

        Parameters
        ----------
        value : ap.Number
            Scale-y value from the center of this instance.

        References
        ----------
        - GraphicsBase scale_x_from_center and scale_y_from_center interfaces
            - https://bit.ly/3ityoCX
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='scale_y_from_center', locals_=locals(),
                module_name=__name__, class_=ScaleYFromCenterInterface):
            from apysc._validation import number_validation
            self._initialize_scale_y_from_center_if_not_initialized()
            number_validation.validate_num(num=value)
            if not isinstance(value, ap.Number):
                value = ap.Number(value)
            before_value: ap.Number = self._scale_y_from_center
            self._scale_y_from_center = value
            self._append_scale_y_from_center_update_expression(
                before_value=before_value)

            self._append_scale_y_from_center_attr_linking_setting()

    def _append_scale_y_from_center_update_expression(
            self, *, before_value: Number) -> None:
        """
        Append the scale-y from the center of this instance
        updating expression.

        Parameters
        ----------
        before_value : ap.Number
            Before updating value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_scale_y_from_center_update_expression,
                locals_=locals(),
                module_name=__name__, class_=ScaleYFromCenterInterface):
            from apysc._type import value_util
            before_value_str: str = value_util.get_value_str_for_expression(
                value=before_value)
            after_value_str: str = value_util.get_value_str_for_expression(
                value=self._scale_y_from_center)
            expression: str = (
                f'{self.variable_name}.scale(1, 1 / {before_value_str});'
                f'\n{self.variable_name}.scale(1, {after_value_str});'
                f'\n{before_value_str} = {after_value_str};'
            )
            ap.append_js_expression(expression=expression)

    _scale_y_from_center_snapshots: Dict[str, float]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_scale_y_from_center_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_scale_y_from_center_snapshots',
            value=self._scale_y_from_center._value,
            snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._scale_y_from_center._value = \
            self._scale_y_from_center_snapshots[snapshot_name]



"""The utilities module for each lint's hash file (used to check whether
the files are updated or not).

Mainly following interfaces are defined:

- get_lint_hash_dir_path
    - Get a specified lint type's hash directory path.
- get_target_module_hash_file_path
    - Get a specified module's hash file path.
- read_target_module_hash
    - Read a specified module's hashed string.
- read_saved_hash
    - Read an already-saved module's hashed string.
- save_target_module_hash
    - Save a target module's current hash.
- save_target_modules_hash
    - Save target modules' current hash.
- is_module_updated
    - Get a boolean value whether a specified module has been updated
        after the last lint execution.
- remove_not_updated_module_paths
    - Remove not updated modules from specified module paths.
"""

import hashlib
import os
from concurrent import futures
from enum import Enum
from multiprocessing import Pool
from multiprocessing import cpu_count
from typing import List

from typing_extensions import TypedDict


class LintType(Enum):

    AUTOFLAKE = 'autoflake'
    ISORT = 'isort'
    AUTOPEP8 = 'autopep8'


_LINT_PACKAGE_ROOT_PATH: str = './.lint_hash/'


def get_lint_hash_dir_path(*, lint_type: LintType) -> str:
    """
    Get a specified lint type's hash directory path.

    Parameters
    ----------
    lint_type : LintType
        Target lint type.

    Returns
    -------
    dir_path : str
        Target lint type's hash directory path.

    Notes
    -----
    Returned directory path will create automatically if it does
    not exist.
    """
    dir_path: str = os.path.join(
        _LINT_PACKAGE_ROOT_PATH,
        f'.{lint_type.value}/'
    )
    os.makedirs(dir_path, exist_ok=True)
    return dir_path


def get_target_module_hash_file_path(
        *, module_path: str, lint_type: LintType) -> str:
    """
    Get a specified module's hash file path.

    Parameters
    ----------
    module_path : str
        Target module path.
    lint_type : LintType
        Target lint type.

    Returns
    -------
    file_path : str
        Target hash file path.

    Notes
    -----
    Returned file's directory path will create automatically if it
    does not exist.
    """
    if module_path.startswith('./'):
        module_path = module_path.replace('./', '', 1)
    dir_path: str = get_lint_hash_dir_path(lint_type=lint_type)
    file_path: str = os.path.join(dir_path, module_path)
    file_path = file_path.replace('.py', '', 1)
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    return file_path


def read_target_module_hash(*, module_path: str) -> str:
    """
    Read a specified module's hashed string.

    Parameters
    ----------
    module_path : str
        Target module path.

    Returns
    -------
    hashed_string : str
        Hashed module string. If there is no module at the specified
        path, then a blank string will be returned.
    """
    from apysc._file import file_util
    if not os.path.isfile(module_path):
        return ''
    module_txt: str = file_util.read_txt(file_path=module_path)
    hashed_string: str = hashlib.sha1(module_txt.encode()).hexdigest()
    return hashed_string


def read_saved_hash(*, module_path: str, lint_type: LintType) -> str:
    """
    Read an already-saved module's hashed string.

    Parameters
    ----------
    module_path : str
        Target module path.
    lint_type : LintType
        Target lint type.

    Returns
    -------
    saved_hash : str
        An already-saved module's hash string. If there is no saved
        hash file then a blank string will be returned.
    """
    from apysc._file import file_util
    file_path: str = get_target_module_hash_file_path(
        module_path=module_path, lint_type=lint_type)
    if not os.path.isfile(file_path):
        return ''
    saved_hash: str = file_util.read_txt(file_path=file_path)
    saved_hash = saved_hash.strip()
    return saved_hash


def save_target_module_hash(*, module_path: str, lint_type: LintType) -> None:
    """
    Save a target module's current hash.

    Parameters
    ----------
    module_path : str
        Target module path.
    lint_type : LintType
        Target lint type.
    """
    from apysc._file import file_util
    hash: str = read_target_module_hash(module_path=module_path)
    if hash == '':
        return
    file_path: str = get_target_module_hash_file_path(
        module_path=module_path, lint_type=lint_type)
    file_util.save_plain_txt(txt=hash, file_path=file_path)


def save_target_modules_hash(
        *, module_paths: List[str], lint_type: LintType) -> None:
    """
    Save target modules' current hash.

    Parameters
    ----------
    module_paths : list of str
        Target module paths.
    lint_type : LintType
        Target lint type.
    """
    workers: int = max(cpu_count() - 1, 1)
    with futures.ProcessPoolExecutor(max_workers=workers) as executor:
        future_list: List[futures.Future] = []
        for module_path in module_paths:
            future = executor.submit(
                save_target_module_hash,
                module_path=module_path,
                lint_type=lint_type)
            future_list.append(future)
        _ = futures.as_completed(fs=future_list)


def is_module_updated(*, module_path: str, lint_type: LintType) -> bool:
    """
    Get a boolean value whether a specified module has been updated
    after the last lint execution.

    Parameters
    ----------
    module_path : str
        Target module path.
    lint_type : LintType
        Target lint type.

    Returns
    -------
    result : bool
        If a specified module has been updated then True will
        be returned.
    """
    saved_hash: str = read_saved_hash(
        module_path=module_path, lint_type=lint_type)
    current_hash: str = read_target_module_hash(module_path=module_path)
    if saved_hash == current_hash:
        return False
    return True


class _IsModuleUpdatedArgs(TypedDict):
    module_path: str
    lint_type: LintType


def _is_module_updated_func_for_multiprocessing(
        args: _IsModuleUpdatedArgs) -> bool:
    """
    Wrapper function of the `is_module_updated` function
    for the multiprocessing.

    Parameters
    ----------
    args : _IsModuleUpdatedArgs
        Arguments dictionary to pass to the `is_module_updated`
        function.

    Returns
    -------
    result : bool
        If a specified module has been updated then True will
        be returned.
    """
    result: bool = is_module_updated(
        module_path=args['module_path'],
        lint_type=args['lint_type'])
    return result


def _create_args_list_for_multiprocessing(
        *, module_paths: List[str],
        lint_type: LintType) -> List[_IsModuleUpdatedArgs]:
    """
    Create an arguments list for the multiprocessing.

    Parameters
    ----------
    module_paths : list of str
        Target Python module paths.
    lint_type : LintType
        Target lint type.

    Returns
    -------
    args_list : list of _IsModuleUpdatedArgs
        Created arguments list for the multiprocessing.
    """
    args_list: List[_IsModuleUpdatedArgs] = []
    for module_path in module_paths:
        args_list.append({
            'module_path': module_path,
            'lint_type': lint_type,
        })
    return args_list


def remove_not_updated_module_paths(
        *, module_paths: List[str],
        lint_type: LintType) -> List[str]:
    """
    Remove not updated modules from specified module paths.

    Parameters
    ----------
    module_paths : list of str
        Target Python module paths.
    lint_type : LintType
        Target lint type.

    Returns
    -------
    sliced_module_paths : list of str
        After the slicing module paths.
    """
    workers: int = max(cpu_count() - 1, 1)
    args_list: List[_IsModuleUpdatedArgs] = \
        _create_args_list_for_multiprocessing(
            module_paths=module_paths, lint_type=lint_type)
    sliced_module_paths: List[str] = []
    with Pool(processes=workers) as p:
        module_updated_bool_list: List[bool] = p.map(
            func=_is_module_updated_func_for_multiprocessing,
            iterable=args_list,
        )
    module_updated: bool
    for i, module_updated in enumerate(module_updated_bool_list):
        if not module_updated:
            continue
        sliced_module_paths.append(module_paths[i])
    return sliced_module_paths



"""Class implementation for the blank object interface.
This interface will be used by the custom event interface
or something else.
"""


class BlankObjectInterface:

    _is_blank_object_initialized: bool = False
    _blank_object_variable_name: str

    def _initialize_blank_object_if_not_initialized(self) -> None:
        """
        Initialize a blank object value if it hasn't been
        initialized yet.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._initialize_blank_object_if_not_initialized,
                locals_=locals(),
                module_name=__name__, class_=BlankObjectInterface):
            from apysc._expression import expression_data_util
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._expression.event_handler_scope import \
                TemporaryNotHandlerScope
            if self._is_blank_object_initialized:
                return
            with TemporaryNotHandlerScope():
                self._blank_object_variable_name = expression_variables_util.\
                    get_next_variable_name(type_name=var_names.BLANK_OBJECT)
                expression: str = (
                    f'var {self._blank_object_variable_name} = {{}};'
                )
                expression_data_util.append_js_expression(
                    expression=expression)
                self._is_blank_object_initialized = True

    @property
    def blank_object_variable_name(self) -> str:
        """
        Get a blank object variable name.

        Returns
        -------
        blank_object_variable_name : str
            A blank object variable name.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='blank_object_variable_name', locals_=locals(),
                module_name=__name__, class_=BlankObjectInterface):
            self._initialize_blank_object_if_not_initialized()
            return self._blank_object_variable_name



"""Class implementation for the animation_time interface.
"""

from apysc._type.number import Number
from apysc._type.variable_name_interface import VariableNameInterface


class AnimationTimeInterface(VariableNameInterface):

    def animation_time(self) -> Number:
        """
        Get an animation elapsed milisecond.

        Returns
        -------
        elapsed_time : Number
            An animation elapsed milisecond.

        References
        ----------
        - animation_time interface document
            - https://simon-ritchie.github.io/apysc/animation_time.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.animation_time, locals_=locals(),
                module_name=__name__, class_=AnimationTimeInterface):
            elapsed_time: Number = Number(0.0)
            expression: str = (
                f'{elapsed_time.variable_name} = '
                f'{self.variable_name}.timeline().time();'
            )
            ap.append_js_expression(expression=expression)
            return elapsed_time



"""Class implementation for the timer event.
"""

from apysc._event.event import Event
from apysc._time import timer


class TimerEvent(Event):
    """
    Timer event class.

    References
    ----------
    - TimerEvent class document
        - https://simon-ritchie.github.io/apysc/timer_event.html
    """

    _this: 'timer.Timer'

    def __init__(self, this: 'timer.Timer') -> None:
        """
        Timer event class.

        Parameters
        ----------
        this : Timer
            Target timer instance.

        References
        ----------
        - TimerEvent class document
            - https://simon-ritchie.github.io/apysc/timer_event.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=TimerEvent):
            from apysc._expression import var_names
            super(TimerEvent, self).__init__(
                this=this, type_name=var_names.TIMER_EVENT)

    @property
    def this(self) -> 'timer.Timer':
        """
        Get a timer instance that listening this event.

        Returns
        -------
        this : TImer
            Instance that listening this event.

        References
        ----------
        - TimerEvent class document
            - https://simon-ritchie.github.io/apysc/timer_event.html
        """
        return self._this



"""Implementation of the Ellipse class.
"""

from typing import Union

from apysc._display import graphics
from apysc._display.cx_interface import CxInterface
from apysc._display.cy_interface import CyInterface
from apysc._display.line_base import LineBase
from apysc._display.width_and_height_interfaces_for_ellipse import \
    WidthAndHeightInterfacesForEllipse
from apysc._type.int import Int


class Ellipse(  # type: ignore
        CxInterface, CyInterface, LineBase,
        WidthAndHeightInterfacesForEllipse):
    """
    The ellipse vector graphics class.

    References
    ----------
    - Graphics draw_ellipse interface
        - https://bit.ly/3xPVicP
    """

    def __init__(
            self,
            *,
            parent: 'graphics.Graphics',
            x: Union[int, Int],
            y: Union[int, Int],
            width: Union[int, Int],
            height: Union[int, Int]) -> None:
        """
        Create a ellipse vector graphics.

        Parameters
        ----------
        parent : Graphics
            Graphics instance to link this graphics.
        x : int or Int
            X-coordinate of the ellipse center.
        y : int or Int
            Y-coordinate of the ellipse center.
        width : int or Int
            Ellipse width.
        height : int or Int
            Ellipse height.

        References
        ----------
        - Graphics draw_ellipse interface
            - https://bit.ly/3xPVicP
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Ellipse):
            from apysc._converter.to_apysc_val_from_builtin import \
                get_copied_int_from_builtin_val
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._validation import size_validation
            variable_name: str = expression_variables_util.\
                get_next_variable_name(type_name=var_names.ELLIPSE)
            super(Ellipse, self).__init__(
                parent=parent, x=0, y=0, variable_name=variable_name)
            size_validation.validate_size_is_gt_zero(size=width)
            size_validation.validate_size_is_gt_zero(size=height)
            self._width = get_copied_int_from_builtin_val(integer=width)
            self._height = get_copied_int_from_builtin_val(integer=height)
            self._set_initial_basic_values(parent=parent)
            self._append_constructor_expression()
            self.x = get_copied_int_from_builtin_val(integer=x)
            self.y = get_copied_int_from_builtin_val(integer=y)
            self._set_line_setting_if_not_none_value_exists(
                parent_graphics=parent)
            self._set_overflow_visible_setting()

            self._append_width_attr_linking_setting()
            self._append_height_attr_linking_setting()

    def _append_constructor_expression(self) -> None:
        """
        Append a constructor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=Ellipse):
            from apysc._display.stage import get_stage_variable_name
            from apysc._type import value_util
            stage_variable_name: str = get_stage_variable_name()
            width_str: str = value_util.get_value_str_for_expression(
                value=self._width)
            height_str: str = value_util.get_value_str_for_expression(
                value=self._height)
            expression: str = (
                f'var {self.variable_name} = {stage_variable_name}'
                f'\n  .ellipse({width_str}, {height_str})'
                '\n  .attr({'
            )
            expression = self._append_basic_vals_expression(
                expression=expression, indent_num=2)
            expression += '\n  });'
            ap.append_js_expression(expression=expression)

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Ellipse('<variable_name>')`).
        """
        repr_str: str = f"Ellipse('{self.variable_name}')"
        return repr_str
